package ru.dbotthepony.mc.otm.container

import com.google.common.collect.ImmutableSet
import com.mojang.datafixers.util.Either
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.MapCodec
import net.minecraft.core.component.DataComponentPatch
import net.minecraft.core.registries.BuiltInRegistries
import net.minecraft.core.registries.Registries
import net.minecraft.tags.TagKey
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import net.neoforged.bus.api.IEventBus
import ru.dbotthepony.mc.otm.util.isNotEmpty
import ru.dbotthepony.mc.otm.registry.MBuiltInRegistries
import ru.dbotthepony.mc.otm.registry.MDeferredRegister
import ru.dbotthepony.mc.otm.registry.MRegistries
import java.util.function.Predicate
import kotlin.jvm.optionals.getOrElse

interface ItemFilter : Predicate<ItemStack> {
	interface Type<T : ItemFilter> {
		val codec: MapCodec<T>
	}

	val type: Type<*>

	/**
	 * Whenever [test] will return `true` no matter the argument,
	 * effectively telling that this filter is "allow all" / "no filter specified"
	 *
	 * Can be treated as (but is not equal to) "isEmpty"
	 */
	val allowAll: Boolean
		get() = false

	/**
	 * Whenever [test] will return `false` no matter the argument,
	 * effectively telling that this filter is "deny all" / "forbidden for automation"
	 */
	val denyAll: Boolean
		get() = false

	/**
	 * Whenever this filter has meaningful rules behind it, e.g. [test] will return either `true` or `false`,
	 * depending on value passed.
	 *
	 * In other words, returns whenever [denyAll] and [allowAll] are both `false`.
	 */
	val hasRules: Boolean
		get() = !denyAll && !allowAll

	val depth: Int
		get() = 1

	val displayItems: ImmutableSet<ItemStackKey>
		get() = ImmutableSet.of()

	private data class Item(val item: net.minecraft.world.item.Item) : ItemFilter {
		override val type: Type<*>
			get() = Companion

		override val denyAll: Boolean
			get() = item === Items.AIR

		override fun test(t: ItemStack): Boolean {
			return t.isNotEmpty && t.item === item
		}

		override val displayItems: ImmutableSet<ItemStackKey>
			get() = ImmutableSet.of(ItemStackKey(item))

		companion object : Type<Item> {
			override val codec: MapCodec<Item> by lazy {
				BuiltInRegistries.ITEM.byNameCodec().xmap(::Item, Item::item).fieldOf("item")
			}
		}
	}

	private data class Tag(val tag: TagKey<net.minecraft.world.item.Item>) : ItemFilter {
		override val type: Type<*>
			get() = Companion

		override fun test(t: ItemStack): Boolean {
			return t.`is`(tag)
		}

		// TODO: can not be "lazy" cached because this will break with /reload command
		override val displayItems: ImmutableSet<ItemStackKey> get() {
			return BuiltInRegistries.ITEM
				.getTag(tag)
				.map { it.stream().map { ItemStackKey(it.value()) }.collect(ImmutableSet.toImmutableSet()) }
				.orElseGet { ImmutableSet.of() }
		}

		companion object : Type<Tag> {
			override val codec: MapCodec<Tag> by lazy {
				TagKey.codec(Registries.ITEM).xmap(::Tag, Tag::tag).fieldOf("tag")
			}
		}
	}

	private object DenyAll : ItemFilter, Type<DenyAll> {
		override val type: Type<*>
			get() = this

		override val codec: MapCodec<DenyAll> = MapCodec.unit(this)

		override val denyAll: Boolean
			get() = true

		override fun test(t: ItemStack): Boolean {
			return false
		}
	}

	companion object : ItemFilter, Type<Companion> {
		const val MAX_DEPTH = 16

		private fun roll(input: ItemFilter): Either<ItemFilter.Item, ItemFilter> {
			if (input is Item) {
				return Either.left(input)
			} else if (input is ItemStackKey && input.components == DataComponentPatch.EMPTY) {
				return Either.left(Item(input.item))
			} else {
				return Either.right(input)
			}
		}

		val CODEC: Codec<ItemFilter> by lazy {
			val codecA = BuiltInRegistries.ITEM
				.byNameCodec()
				.xmap(::Item, Item::item)

			val codecB = MBuiltInRegistries.ITEM_FILTER
				.byNameCodec()
				.dispatch(ItemFilter::type, { it.codec })

			Codec.either(codecA, codecB)
				.xmap({ it.right().getOrElse { it.left().get() } }, ::roll)
				.validate {
					if (it.depth <= MAX_DEPTH) {
						return@validate DataResult.success(it)
					} else {
						return@validate DataResult.error { "Too deep item filter, max depth of $MAX_DEPTH is allowed (depth: ${it.depth})" }
					}
				}
		}

		val EMPTY: ItemFilter
			get() = this

		val DENY_ALL: ItemFilter
			get() = DenyAll

		private val registrar = MDeferredRegister(MRegistries.ITEM_FILTER)

		init {
			registrar.register("item") { Item.Companion }
			registrar.register("item_stack") { ItemStackKey.Companion }
			registrar.register("tag") { Tag.Companion }
			registrar.register("set") { ItemFilterSet.Companion }
			registrar.register("empty") { this }
			registrar.register("deny_all") { DenyAll }
		}

		internal fun register(bus: IEventBus) {
			registrar.register(bus)
		}

		override val type: Type<*>
			get() = this

		override fun test(t: ItemStack): Boolean {
			return true
		}

		override val codec: MapCodec<Companion> = MapCodec.unit(this)

		override val allowAll: Boolean
			get() = true

		@JvmStatic
		fun item(item: net.minecraft.world.item.Item): ItemFilter {
			return Item(item)
		}

		@JvmStatic
		fun item(item: ItemStack): ItemFilter {
			return Item(item.item)
		}

		@JvmStatic
		fun itemAndComponents(item: ItemStack): ItemFilter {
			return ItemStackKey(item)
		}

		@JvmStatic
		fun tag(tag: TagKey<net.minecraft.world.item.Item>): ItemFilter {
			return Tag(tag)
		}
	}
}
