package ru.dbotthepony.mc.otm.container

import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.capability.IMatteryUpgrade
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.UpgradeType
import ru.dbotthepony.mc.otm.config.EnergyBalanceValues
import ru.dbotthepony.mc.otm.config.VerboseEnergyBalanceValues
import ru.dbotthepony.kommons.collect.map
import ru.dbotthepony.kommons.collect.reduce
import ru.dbotthepony.mc.otm.util.math.Decimal
import java.util.function.BooleanSupplier
import kotlin.math.pow

class UpgradeContainer(
	slotCount: Int,
	val allowedUpgrades: Set<UpgradeType> = UpgradeType.ALL,
	val shouldLockUpgradeSlots: BooleanSupplier = BooleanSupplier { false },
	private val listener: Runnable = Runnable {}
) : EnhancedContainer.Simple(slotCount), IMatteryUpgrade {
	override fun notifySlotChanged(slot: Int, old: ItemStack) {
		super.notifySlotChanged(slot, old)
		listener.run()
	}

	override val upgradeTypes: Set<UpgradeType>
		get() = setOf()

	private inline fun positiveDecimals(fn: (IMatteryUpgrade) -> Decimal, reducer: (Decimal, Decimal) -> Decimal): Decimal {
		if (isEmpty)
			return Decimal.ZERO

		var result = Decimal.ZERO

		for (it in iterator()) {
			val cap = it.getCapability(MatteryCapability.UPGRADE) ?: continue
			val value = fn(cap)

			if (value > Decimal.ZERO) {
				if (it.count == 1) {
					result = reducer(result, value )
				} else {
					result = reducer(result, value * it.count)
				}
			}
		}

		return result
	}

	private inline fun anyDecimals(fn: (IMatteryUpgrade) -> Decimal, reducer: (Decimal, Decimal) -> Decimal): Decimal {
		if (isEmpty)
			return Decimal.ZERO

		var result = Decimal.ZERO

		for (it in iterator()) {
			val cap = it.getCapability(MatteryCapability.UPGRADE) ?: continue
			val value = fn(cap)

			if (it.count == 1) {
				result = reducer(result, value)
			} else {
				result = reducer(result, value * it.count)
			}
		}

		return result
	}

	override val speedBonus: Double get() {
		if (isEmpty)
			return 0.0

		return sumOf { (it.getCapability(MatteryCapability.UPGRADE)?.speedBonus ?: 0.0) * it.count }
	}

	override val processingItems: Int get() {
		if (isEmpty)
			return 0

		return sumOf { (it.getCapability(MatteryCapability.UPGRADE)?.processingItems ?: 0).coerceAtLeast(0) * it.count }
	}

	override val energyStorageFlat: Decimal
		get() = positiveDecimals(IMatteryUpgrade::energyStorageFlat, Decimal::plus)
	override val energyStorage: Decimal
		get() = positiveDecimals(IMatteryUpgrade::energyStorage, Decimal::plus)
	override val matterStorageFlat: Decimal
		get() = positiveDecimals(IMatteryUpgrade::matterStorageFlat, Decimal::plus)
	override val matterStorage: Decimal
		get() = positiveDecimals(IMatteryUpgrade::matterStorage, Decimal::plus)
	override val energyConsumed: Decimal
		get() = anyDecimals(IMatteryUpgrade::energyConsumed, Decimal::plus)

	override val failureMultiplier: Double get() {
		if (isEmpty)
			return 1.0

		var result = 1.0

		for (it in iterator()) {
			val cap = it.getCapability(MatteryCapability.UPGRADE) ?: continue
			val chance = cap.failureMultiplier.coerceAtLeast(0.0)

			if (chance == 0.0)
				return 0.0
			else if (it.count == 1)
				result *= chance
			else
				result *= chance.pow(it.count)
		}

		return result
	}

	override val energyThroughputFlat: Decimal
		get() = positiveDecimals(IMatteryUpgrade::energyThroughputFlat, Decimal::plus)
	override val energyThroughput: Decimal
		get() = positiveDecimals(IMatteryUpgrade::energyThroughput, Decimal::plus)

	fun transform(values: EnergyBalanceValues): EnergyBalanceValues {
		return object : EnergyBalanceValues {
			override val energyCapacity: Decimal
				get() = if (isEmpty) values.energyCapacity else values.energyCapacity * (energyStorage + Decimal.ONE) + energyStorageFlat
			override val energyThroughput: Decimal
				get() = if (isEmpty) values.energyThroughput else values.energyThroughput * (this@UpgradeContainer.energyThroughput + Decimal.ONE) + energyThroughputFlat
		}
	}

	fun transform(values: VerboseEnergyBalanceValues): VerboseEnergyBalanceValues {
		return object : VerboseEnergyBalanceValues {
			override val energyCapacity: Decimal
				get() = if (isEmpty) values.energyCapacity else values.energyCapacity * (energyStorage + Decimal.ONE) + energyStorageFlat
			override val maxEnergyReceive: Decimal
				get() = if (isEmpty) values.maxEnergyReceive else values.maxEnergyReceive * (energyThroughput + Decimal.ONE) + energyThroughputFlat
			override val maxEnergyExtract: Decimal
				get() =if (isEmpty) values.maxEnergyExtract else  values.maxEnergyExtract * (energyThroughput + Decimal.ONE) + energyThroughputFlat
		}
	}

	fun matterCapacity(value: () -> Decimal): () -> Decimal {
		return { if (isEmpty) value.invoke() else value.invoke() * (matterStorage + Decimal.ONE) + matterStorageFlat }
	}
}
