package ru.dbotthepony.mc.otm.util

import io.netty.handler.codec.EncoderException
import net.minecraft.core.IdMap
import net.minecraft.core.registries.BuiltInRegistries
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.NbtAccounter
import net.minecraft.nbt.NbtIo
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.RegistryFriendlyByteBuf
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.ComponentSerialization
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.util.math.Decimal
import java.io.*
import java.math.BigDecimal
import java.math.BigInteger
import kotlin.enums.EnumEntries
import kotlin.math.absoluteValue

fun OutputStream.writeInt(value: Int) {
	if (this is DataOutput) {
		writeInt(value)
		return
	}

	write(value ushr 24)
	write(value ushr 16)
	write(value ushr 8)
	write(value)
}

fun InputStream.readInt(): Int {
	if (this is DataInput) {
		return readInt()
	}

	return (read() shl 24) or (read() shl 16) or (read() shl 8) or read()
}

fun OutputStream.writeLong(value: Long) {
	if (this is DataOutput) {
		writeLong(value)
		return
	}

	write((value ushr 48).toInt())
	write((value ushr 40).toInt())
	write((value ushr 32).toInt())
	write((value ushr 24).toInt())
	write((value ushr 16).toInt())
	write((value ushr 8).toInt())
	write(value.toInt())
}

fun InputStream.readLong(): Long {
	if (this is DataInput) {
		return readLong()
	}

	return (read().toLong() shl 48) or
		(read().toLong() shl 40) or
		(read().toLong() shl 32) or
		(read().toLong() shl 24) or
		(read().toLong() shl 16) or
		(read().toLong() shl 8) or
		read().toLong()
}

fun OutputStream.writeFloat(value: Float) = writeInt(value.toBits())
fun InputStream.readFloat() = Float.fromBits(readInt())
fun OutputStream.writeDouble(value: Double) = writeLong(value.toBits())
fun InputStream.readDouble() = Double.fromBits(readLong())

fun InputStream.readVarIntLE(): Int {
	val readFirst = read()

	if (readFirst < 0) {
		throw NoSuchElementException("Reached end of stream")
	}

	if (readFirst and 64 == 0) {
		return if (readFirst and 128 != 0) -(readFirst and 63) else readFirst and 63
	}

	var result = 0
	var nextBit = readFirst and 64
	var read = readFirst and 63
	var i = 0

	while (nextBit != 0) {
		result = result or (read shl i)
		read = read()

		if (read < 0) {
			throw NoSuchElementException("Reached end of stream")
		}

		nextBit = read and 128
		read = read and 127

		if (i == 0)
			i = 6
		else
			i += 7
	}

	result = result or (read shl i)
	return if (readFirst and 128 != 0) -result else result
}

fun OutputStream.writeVarIntLE(value: Int) {
	write((if (value < 0) 128 else 0) or (if (value in -63 .. 63) 0 else 64) or (value.absoluteValue and 63))
	var written = value.absoluteValue ushr 6

	while (written != 0) {
		write((written and 127) or (if (written >= 128) 128 else 0))
		written = written ushr 7
	}
}

fun InputStream.readVarLongLE(sizeLimit: NbtAccounter? = null): Long {
	sizeLimit?.accountBytes(1L)

	val readFirst = read()

	if (readFirst < 0) {
		throw NoSuchElementException("Reached end of stream")
	}

	if (readFirst and 64 == 0) {
		return if (readFirst and 128 != 0) -(readFirst and 63).toLong() else (readFirst and 63).toLong()
	}

	var result = 0L
	var nextBit = readFirst and 64
	var read = readFirst and 63
	var i = 0

	while (nextBit != 0) {
		result = result or (read shl i).toLong()
		sizeLimit?.accountBytes(1L)
		read = read()

		if (read < 0) {
			throw NoSuchElementException("Reached end of stream")
		}

		nextBit = read and 128
		read = read and 127

		if (i == 0)
			i = 6
		else
			i += 7
	}

	result = result or (read shl i).toLong()
	return if (readFirst and 128 != 0) -result else result
}

fun OutputStream.writeVarLongLE(value: Long) {
	write((if (value < 0L) 128 else 0) or (if (value in -63 .. 63) 0 else 64) or (value.absoluteValue and 63).toInt())
	var written = value.absoluteValue ushr 6

	while (written != 0L) {
		write((written and 127).toInt() or (if (written >= 128) 128 else 0))
		written = written ushr 7
	}
}

fun InputStream.readBinaryString(): String {
	val size = readVarIntLE()
	require(size >= 0) { "Negative payload size: $size" }
	val bytes = ByteArray(size)
	read(bytes)
	return bytes.decodeToString()
}

fun OutputStream.writeBinaryString(input: String) {
	val bytes = input.encodeToByteArray()
	writeVarIntLE(bytes.size)
	write(bytes)
}

fun FriendlyByteBuf.readDecimal() = Decimal.read(this)
fun FriendlyByteBuf.writeDecimal(value: Decimal) = value.write(this)

fun InputStream.readDecimal(): Decimal {
	val size = readVarIntLE()
	require(size >= 0) { "Negative payload size: $size" }
	val bytes = ByteArray(size)
	read(bytes)
	return Decimal.fromByteArray(bytes)
}

fun OutputStream.writeDecimal(value: Decimal) {
	val bytes = value.toByteArray()
	writeVarIntLE(bytes.size)
	write(bytes)
}

fun <T : Any> FriendlyByteBuf.writeType(registry: IdMap<T>, value: T) {
	writeInt(registry.getId(value))
}

fun FriendlyByteBuf.writeBlockType(value: Block) {
	writeType(BuiltInRegistries.BLOCK, value)
}

fun FriendlyByteBuf.writeBlockState(value: BlockState) {
	writeVarInt(Block.BLOCK_STATE_REGISTRY.getIdOrThrow(value))
}

fun FriendlyByteBuf.writeItemType(value: Item) {
	writeType(BuiltInRegistries.ITEM, value)
}

fun RegistryFriendlyByteBuf.writeItem(value: ItemStack) {
	ItemStack.OPTIONAL_STREAM_CODEC.encode(this, value)
}

fun RegistryFriendlyByteBuf.writeComponent(value: Component) {
	ComponentSerialization.STREAM_CODEC.encode(this, value)
}

fun OutputStream.writeItemType(value: Item) {
	writeVarIntLE(BuiltInRegistries.ITEM.getId(value))
}

fun <E : Enum<E>> FriendlyByteBuf.readEnum(entries: EnumEntries<E>): E {
	return entries[readVarInt()]
}

fun <T : Any> FriendlyByteBuf.readType(registry: IdMap<T>): T {
	val id = readInt()
	return registry.byId(id) ?: throw NoSuchElementException("No such entry with ID $id")
}

fun FriendlyByteBuf.readBlockType(): Block {
	return readType(BuiltInRegistries.BLOCK)
}

fun FriendlyByteBuf.readBlockState(): BlockState {
	return Block.BLOCK_STATE_REGISTRY.byIdOrThrow(readVarInt())
}

fun FriendlyByteBuf.readItemType(): Item {
	return readType(BuiltInRegistries.ITEM)
}

fun RegistryFriendlyByteBuf.readItem(): ItemStack {
	return ItemStack.OPTIONAL_STREAM_CODEC.decode(this)
}

fun RegistryFriendlyByteBuf.readComponent(): Component {
	return ComponentSerialization.STREAM_CODEC.decode(this)
}

fun InputStream.readItemType(): Item {
	return BuiltInRegistries.ITEM.byId(readVarIntLE())
}

fun FriendlyByteBuf.writeBigInteger(value: BigInteger) {
	writeByteArray(value.toByteArray())
}

fun FriendlyByteBuf.readBigInteger(byteLimit: Int = 128) = BigInteger(readByteArray(byteLimit))
